/*
 * Driver for the Epson RTC module RX-4111 CE
 * 
 * Copyright(C) 2020 VARIKOREA, Sam Kim <sam@varikorea.co.kr>
 */
#include <linux/bcd.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/of_gpio.h>
#include <linux/regmap.h>
#include <linux/rtc.h>
#include <linux/spi/spi.h>

#define DEBUG

/* BANK DEFINE */
#define BANK1_WRITE			(0x01 << 4)
#define BANK2_WRITE			(0x02 << 4) 
#define BANK3_WRITE			(0x03 << 4)
#define BANK6_WRITE			(0x06 << 4)

#define BANK1_READ			(0x09 << 4)
#define BANK2_READ			(0x0A << 4) 
#define BANK3_READ			(0x0B << 4)
#define BANK6_READ			(0x0E << 4)

/* Bank 1 */
#define RX4111_BANK1_SEC		0x00
#define RX4111_BANK1_MIN     		0x01
#define RX4111_BANK1_HOUR    		0x02
#define RX4111_BANK1_WEEK		0x03
#define RX4111_BANK1_DAY		0x04
#define RX4111_BANK1_MONTH   		0x05
#define RX4111_BANK1_YEAR    		0x06
#define RX4111_BANK1_MIN_ALARM		0x07
#define RX4111_BANK1_HOUR_ALARM  	0x08
#define RX4111_BANK1_WEEK_DAY_ALARM  	0x09
#define RX4111_BANK1_TIMER_COUNTER0 	0x0A
#define RX4111_BANK1_TIMER_COUNTER1 	0x0B
#define RX4111_BANK1_TIMER_COUNTER2 	0x0C
#define RX4111_BANK1_EXTENSION_REG	0x0D
#define RX4111_BANK1_FLAG_REG    	0x0E
#define RX4111_BANK1_CTRL_REG		0x0F

/* Bank 2 */
#define RX4111_BANK2_TIMESTAMP_1_1024S	0x00
#define RX4111_BANK2_TIMESTAMP_1_256S	0x01
#define RX4111_BANK2_TIMESTAMP_SEC	0x02
#define RX4111_BANK2_TIMESTAMP_MIN	0x03
#define RX4111_BANK2_TIMESTAMP_HOUR	0x04
#define RX4111_BANK2_TIMESTAMP_WEEK	0x05
#define RX4111_BANK2_TIMESTAMP_DAY	0x06
#define RX4111_BANK2_TIMESTAMP_MONTHS	0x07
#define RX4111_BANK2_TIMESTAMP_YEAR	0x08
#define RX4111_BANK2_STATUS_STAMP	0x09
#define RX4111_BANK2_OVER_WRITE_CTRL	0x0B
#define RX4111_BANK2_SEC_ALARM		0x0C
#define RX4111_BANK2_TIMER_CTRL		0x0D
#define RX4111_BANK2_TIMESTAMP_CTRL0	0x0E
#define RX4111_BANK2_COMMAND_TRIGGER	0x0F

/* Bank 3 */
#define RX4111_BANK3_PWR_SWITCH_CTRL	0x02
#define RX4111_BANK3_TIMESTAMP_CTRL1	0x04
#define RX4111_BANK3_TIMESTAMP_CTRL2	0x05
#define RX4111_BANK3_TIMESTAMP_CTRL3	0x06
#define RX4111_BANK3_TEST		0x0F

/* Bank 4,5,6,7 */
#define RX4111_TIMESTAMP_1_256S_1	0x00
#define RX4111_TIMESTAMP_SEC_1		0x01
#define RX4111_TIMESTAMP_MIN_1		0x02
#define RX4111_TIMESTAMP_HOUR_1		0x03
#define RX4111_TIMESTAMP_DAY_1		0x04
#define RX4111_TIMESTAMP_MONTHS_1	0x05
#define RX4111_TIMESTAMP_YEAR_1		0x06
#define RX4111_STATUS_STAMP_1		0x07

#define RX4111_TIMESTAMP_1_256S_2	0x08
#define RX4111_TIMESTAMP_SEC_2		0x09
#define RX4111_TIMESTAMP_MIN_2		0x0A
#define RX4111_TIMESTAMP_HOUR_2		0x0B
#define RX4111_TIMESTAMP_DAY_2		0x0C
#define RX4111_TIMESTAMP_MONTHS_2	0x0D
#define RX4111_TIMESTAMP_YEAR_2		0x0E
#define RX4111_STATUS_STAMP_2		0x0F

#define ON_DELAY	40

#define RX4111_BIT_ALARM_EN		BIT(7)

/* Extension Register Bank1 (Dh) bit positions */
#define RX4111_BIT_EXT_TSEL0		BIT(0)
#define RX4111_BIT_EXT_TSEL1		BIT(1)
#define RX4111_BIT_EXT_WADA		BIT(3)
#define RX4111_BIT_EXT_TE		BIT(4)
#define RX4111_BIT_EXT_USEL		BIT(5)
#define RX4111_BIT_EXT_FSEL0		BIT(6)
#define RX4111_BIT_EXT_FSEL1		BIT(7)

/* Flag Register Bank1 (Eh) bit positions */
#define RX4111_BIT_FLAG_XST		BIT(0)
#define RX4111_BIT_FLAG_VLF		BIT(1)
#define RX4111_BIT_FLAG_EVF		BIT(2)
#define RX4111_BIT_FLAG_AF		BIT(3)
#define RX4111_BIT_FLAG_TF		BIT(4)
#define RX4111_BIT_FLAG_UF		BIT(5)
#define RX4111_BIT_FLAG_POR		BIT(7)

/* Control Register Bank1 (Fh) bit positions */
#define RX4111_BIT_CTRL_STOP		BIT(0)
#define RX4111_BIT_CTRL_EIE		BIT(2)
#define RX4111_BIT_CTRL_AIE		BIT(3)
#define RX4111_BIT_CTRL_TIE		BIT(4)
#define RX4111_BIT_CTRL_UIE		BIT(5)

/* Power Switch Control Bank3 (2h) bit positions */
#define RX4111_BIT_PSC_SMPT0		BIT(0)
#define RX4111_BIT_PSC_SMPT1		BIT(1)
#define RX4111_BIT_PSC_SWSEL0		BIT(2)
#define RX4111_BIT_PSC_SWSEL1		BIT(3)
#define RX4111_BIT_PSC_INIEN		BIT(6)

enum {
	RTC_SEC = 0,
	RTC_MIN,
	RTC_HOUR,
	RTC_WDAY,
	RTC_MDAY,
	RTC_MONTH,
	RTC_YEAR,
	RTC_NR_TIME
};

#define RX4111_DRIVER_NAME		"rtc-rx4111"

struct rx4111_data {
	struct rtc_device *rtc;
	struct regmap *regmap;
	struct mutex mutex;
	int irq;
};

static int get_week_day( u8 reg_week_day )
{
	int i, tm_wday = -1;

	for ( i=0; i < 7; i++ )
	{
		if ( reg_week_day & 1 )
		{
			tm_wday = i;
			break;
		}
		reg_week_day >>= 1;
	}

	return	tm_wday;
}

static int rx4111_rtc_tm_to_data(struct rtc_time *tm, u8 *data)
{
	/*
	 * The year in the RTC is a value between 0 and 99.
	 * Assume that this represents the current century
	 * and disregard all other values.
	 */
	if (tm->tm_year < 100 || tm->tm_year >= 200)
		return -EINVAL;

	data[RTC_SEC] = bin2bcd(tm->tm_sec);
	data[RTC_MIN] = bin2bcd(tm->tm_min);
	data[RTC_HOUR] = bin2bcd(tm->tm_hour);
	data[RTC_WDAY] = BIT(bin2bcd(tm->tm_wday));
	data[RTC_MDAY] = bin2bcd(tm->tm_mday);
	data[RTC_MONTH] = bin2bcd(tm->tm_mon + 1);
	data[RTC_YEAR] = bin2bcd(tm->tm_year % 100);

	return 0;
}

static int rx4111_data_to_rtc_tm(u8 *data, struct rtc_time *tm)
{
	tm->tm_sec	= bcd2bin(data[0]);
	tm->tm_min	= bcd2bin(data[1]);
	tm->tm_hour	= bcd2bin(data[2] & 0x3F);
	tm->tm_wday	= bcd2bin(data[3]) - 1;
	tm->tm_mday	= bcd2bin(data[4]);
	tm->tm_mon	= bcd2bin(data[5] & 0x1F) - 1;
	tm->tm_year	= bcd2bin(data[6]) + 100;

	if (tm->tm_year < 100 || tm->tm_year >= 200)
		return -EINVAL;

	return 0;
}

static int rx4111_set_time(struct device *dev, struct rtc_time *tm)
{
	struct rx4111_data *rx4111 = dev_get_drvdata(dev);
	u8 data[RTC_NR_TIME];
	int ret;

	/* set STOP bit before changing clock/calendar */
	ret = regmap_update_bits(rx4111->regmap, BANK1_WRITE | RX4111_BANK1_CTRL_REG, RX4111_BIT_CTRL_STOP, RX4111_BIT_CTRL_STOP);
	if (ret)
		return ret;

	ret = rx4111_rtc_tm_to_data(tm, data);
	if (ret < 0)
		return ret;

	ret = regmap_bulk_write(rx4111->regmap, BANK1_WRITE | RX4111_BANK1_SEC, data, RTC_NR_TIME);
	if (ret)
		return ret;

	/* clear STOP bit after changing clock/calendar */
	ret = regmap_update_bits(rx4111->regmap, BANK1_WRITE | RX4111_BANK1_CTRL_REG, RX4111_BIT_CTRL_STOP, 0);

	return ret;
}

static int rx4111_get_time(struct device *dev, struct rtc_time *tm)
{
	struct rx4111_data *rx4111 = dev_get_drvdata(dev);
	u8 data[RTC_NR_TIME];
	int ret;

	/* read registers to date */
	ret = regmap_bulk_read(rx4111->regmap, BANK1_READ | RX4111_BANK1_SEC, data, RTC_NR_TIME);
	if (ret)
		return ret;

	ret = rx4111_data_to_rtc_tm(data, tm);
	if (ret)
		return ret;

	return rtc_valid_tm(tm);
}

static int rx4111_set_alarm(struct device *dev, struct rtc_wkalrm *alarm)
{
	struct rx4111_data *rx4111 = dev_get_drvdata(dev);
	int ret;
	u8 bank1_data[3];
	u8 bank2_data;
	unsigned int ctrlreg;

	if (rx4111->irq <= 0)
		return -EINVAL;

	/* AIE bit "0" */
	ret = regmap_update_bits(rx4111->regmap,BANK1_WRITE | RX4111_BANK1_CTRL_REG, RX4111_BIT_CTRL_AIE , 0);
	if (ret)
		return ret;

	bank2_data    = bin2bcd(alarm->time.tm_sec);
	bank1_data[0] = bin2bcd(alarm->time.tm_min);
	bank1_data[1] = bin2bcd(alarm->time.tm_hour);
	bank1_data[2] = bin2bcd(alarm->time.tm_mday);

	ret = regmap_bulk_write(rx4111->regmap,BANK1_WRITE | RX4111_BANK1_MIN_ALARM, bank1_data, 3);
	if (ret)
		return ret;

	ret = regmap_write(rx4111->regmap, BANK2_WRITE | RX4111_BANK2_SEC_ALARM, bank2_data);
	if (ret)
		return ret;

	/* Select week or day in WADA bit */
	/* set to Day Alarm */
	ret = regmap_update_bits(rx4111->regmap,BANK1_WRITE | RX4111_BANK1_EXTENSION_REG, RX4111_BIT_EXT_WADA, 1);
	if (ret)
		return ret;

	/* clear AF bit */
	ret = regmap_update_bits(rx4111->regmap,BANK1_WRITE | RX4111_BANK1_FLAG_REG, RX4111_BIT_FLAG_AF, 0);
	if (ret)
		return ret;

	//re-enable interrupt if required
	if (alarm->enabled) {

		if ( rx4111->rtc->uie_rtctimer.enabled )
			ctrlreg |= RX4111_BIT_CTRL_UIE;				//set update interrupt enable
		if ( rx4111->rtc->aie_timer.enabled )
			ctrlreg |= (RX4111_BIT_CTRL_AIE | RX4111_BIT_CTRL_UIE);	//set alarm interrupt enable

		ret = regmap_write(rx4111->regmap, BANK1_WRITE | RX4111_BANK1_CTRL_REG, ctrlreg);
		if (ret)
			return ret;
	}
	return 0;
}

static int rx4111_read_alarm(struct device *dev, struct rtc_wkalrm *alarm)
{
	struct rx4111_data *rx4111 = dev_get_drvdata(dev);
	u8 alarmvals[3];		//minute, hour, week/day values
	unsigned int ext, flag, ctrl;
	int ret = 0;

	if (rx4111->irq <= 0)
		return -EINVAL;

	//get current minute, hour, week/day alarm values
	ret = regmap_bulk_read(rx4111->regmap,BANK1_READ | RX4111_BANK1_MIN_ALARM, alarmvals, 3);
	if (ret)
		return ret;

	ret = regmap_read(rx4111->regmap, BANK1_READ | RX4111_BANK1_CTRL_REG, &ctrl);
	if (ret)
                return ret;

	ret = regmap_read(rx4111->regmap, BANK1_READ | RX4111_BANK1_FLAG_REG, &flag);
	if (ret)
                return ret;

	ret = regmap_read(rx4111->regmap, BANK1_READ | RX4111_BANK1_EXTENSION_REG, &ext);
	if (ret)
                return ret;

	// Hardware alarm precision is 1 minute
	alarm->time.tm_sec = 0;
	//0x7f filters AE bit currently
	alarm->time.tm_min = bcd2bin(alarmvals[0] & 0x7f);		
	//0x3f filters AE bit currently, also 24hr only
	alarm->time.tm_hour = bcd2bin(alarmvals[1] & 0x3f);		
	if ((ext & RX4111_BIT_EXT_WADA))
	{
		alarm->time.tm_mday = (alarmvals[2]&0x80) ? -1 : bcd2bin(alarmvals[2] & 0x3f);	
		alarm->time.tm_wday = -1;
	}	
	else
	{
		alarm->time.tm_mday = -1;
		alarm->time.tm_wday = (alarmvals[2]&0x80) ? -1 : get_week_day(alarmvals[2] & 0x7f);			
	}
	alarm->time.tm_mon = -1;
	alarm->time.tm_year = -1;

	alarm->enabled = !!(ctrl & RX4111_BIT_CTRL_AIE);		//check if interrupt is enabled
	alarm->pending = (flag & RX4111_BIT_FLAG_AF) && alarm->enabled;	//check if flag is triggered

	return ret;
}

static irqreturn_t rx4111_irq_handler(int irq, void *dev_id)
{
	struct device *dev = dev_id;
	struct rx4111_data *rx4111 = dev_get_drvdata(dev);
	struct mutex *lock = &rx4111->rtc->ops_lock;
	unsigned int flags;
	int ret;

	mutex_lock(lock);

	ret = regmap_read(rx4111->regmap, BANK1_READ | RX4111_BANK1_FLAG_REG , &flags);
	if (ret) {
		dev_warn(dev, "Read IRQ Flag Register error %d\n", ret);
		mutex_unlock(lock);
		return IRQ_NONE;
	}

	// check VLF
	if ((flags & RX4111_BIT_FLAG_VLF))
		dev_warn(dev, "Frequency stop was detected, probably due to a supply voltage drop\n");

	// periodic "fixed-cycle" timer
	if (flags & RX4111_BIT_FLAG_TF) {
		flags &= ~RX4111_BIT_FLAG_TF;
		local_irq_disable();
		rtc_update_irq(rx4111->rtc, 1, RTC_PF | RTC_IRQF);
		local_irq_enable();
	}

	// alarm function
	if (flags & RX4111_BIT_FLAG_AF) {
		flags &= ~RX4111_BIT_FLAG_AF;
		local_irq_disable();
		rtc_update_irq(rx4111->rtc, 1, RTC_AF | RTC_IRQF);
		local_irq_enable();
	}

	// time update function
	if (flags & RX4111_BIT_FLAG_UF) {
		flags &= ~RX4111_BIT_FLAG_UF;
		local_irq_disable();
		rtc_update_irq(rx4111->rtc, 1, RTC_UF | RTC_IRQF);
		local_irq_enable();
	}

	// acknowledge IRQ (clear flags)
	ret = regmap_write(rx4111->regmap, BANK1_WRITE | RX4111_BANK1_FLAG_REG, flags);
	if (ret)
		return ret;

	mutex_unlock(lock);

	return IRQ_HANDLED;
}

static int rx4111_alarm_irq_enable(struct device *dev, unsigned int enabled)
{
	struct rx4111_data *rx4111 = dev_get_drvdata(dev);
	int ret;
	unsigned int ctrl, flag;
	u8 status;

	if (rx4111->irq <= 0)
		return -EINVAL;

	ret = regmap_read(rx4111->regmap, BANK1_READ | RX4111_BANK1_CTRL_REG, &ctrl);
	if (ret)
		return ret;

	if (enabled)
	{
		//set update interrupt enable
		if ( rx4111->rtc->uie_rtctimer.enabled )
			status |= RX4111_BIT_CTRL_UIE;						
		//set alarm interrupt enable
		if ( rx4111->rtc->aie_timer.enabled )
			status |= RX4111_BIT_CTRL_AIE | RX4111_BIT_CTRL_UIE;				
	}
	else
	{
		//clear update interrupt enable
		if ( ! rx4111->rtc->uie_rtctimer.enabled )
			status &= ~RX4111_BIT_CTRL_UIE;
		//clear alarm interrupt enable
		if ( ! rx4111->rtc->aie_timer.enabled )
		{
			if ( rx4111->rtc->uie_rtctimer.enabled )
				status &= ~RX4111_BIT_CTRL_AIE;
			else
				status &= ~(RX4111_BIT_CTRL_AIE | RX4111_BIT_CTRL_UIE);
		}		
	}

	//clear alarm flag
	ret = regmap_read(rx4111->regmap, BANK1_READ | RX4111_BANK1_FLAG_REG, &flag);
	if (ret)
		return ret;

	flag &= ~RX4111_BIT_FLAG_AF;
	ret = regmap_write(rx4111->regmap, BANK1_WRITE | RX4111_BANK1_FLAG_REG, flag);
	if (ret)
		return ret;

	//update the Control register if the setting changed
	if (status != ctrl) {
		ctrl = status;
		ret = regmap_write(rx4111->regmap, BANK1_WRITE | RX4111_BANK1_CTRL_REG, ctrl);
		if (ret)
			return ret;
	}
	return 0;
}
static int rx4111_init(struct rx4111_data *rx4111)
{
	unsigned int flags;
	int ret;

	msleep(40);

	ret = regmap_read(rx4111->regmap, BANK1_READ | RX4111_BANK1_FLAG_REG, &flags);
	if (ret)
		return ret;

	/* check for VLF Flag (set at power-on) */
	if ((flags & RX4111_BIT_FLAG_VLF)){
		while(1){
			/* clear flags VLF */
			ret = regmap_update_bits(rx4111->regmap, BANK1_WRITE | RX4111_BANK1_FLAG_REG, RX4111_BIT_FLAG_VLF, 0);
			msleep(200);

			ret = regmap_read(rx4111->regmap, BANK1_READ | RX4111_BANK1_FLAG_REG, &flags);
			if (ret)
				return ret;
			if ((flags == 0))	//VLF=0?
			{
				//initialize
				ret = regmap_update_bits(rx4111->regmap, BANK3_WRITE | RX4111_BANK3_PWR_SWITCH_CTRL, RX4111_BIT_PSC_SWSEL0, 1);
				if (ret)
					return ret;

				ret = regmap_update_bits(rx4111->regmap, BANK1_WRITE | RX4111_BANK1_EXTENSION_REG, RX4111_BIT_EXT_TSEL1, 1);
				if (ret)
					return ret;

				ret = regmap_write(rx4111->regmap,BANK1_WRITE| RX4111_BANK1_FLAG_REG, 0);
				if (ret)
					return ret;

				ret = regmap_update_bits(rx4111->regmap,RX4111_BANK1_CTRL_REG, RX4111_BIT_CTRL_AIE | RX4111_BIT_CTRL_UIE | RX4111_BIT_CTRL_TIE, 0);
				if (ret)
					return ret;
			}
		}
	}
	return ret;
}
static struct rtc_class_ops rx4111_rtc_ops = {
	.read_time = rx4111_get_time,
	.set_time = rx4111_set_time,
	.read_alarm = rx4111_read_alarm,
	.set_alarm = rx4111_set_alarm,
	.alarm_irq_enable = rx4111_alarm_irq_enable,
};
struct regmap_config regmap_spi_config = {
	.reg_bits = 8,
	.val_bits = 8,
};

static void rx4111_dev_info_check(struct spi_device *spi)
{
	/* dev info check */
	if ((spi->bits_per_word && spi->bits_per_word != 8) ||
			(spi->max_speed_hz < 4000000) ||
			(spi->mode != (SPI_CS_HIGH | SPI_CPOL | SPI_CPHA))) {	//rx4111 Mode3
		dev_warn(&spi->dev, "SPI settings: bits_per_word: %d, max_speed_hz: %d, mode: %xh\n",
				spi->bits_per_word, spi->max_speed_hz, spi->mode);
		dev_warn(&spi->dev, "driving device in an unsupported mode");
	}
	dev_info(&spi->dev, "SPI settings: bits_per_word: %d, max_speed_hz: %d, mode: %xh\n",
				spi->bits_per_word, spi->max_speed_hz, spi->mode);
}

static int rx4111_probe(struct spi_device *spi)
{
	struct rx4111_data *rx4111;
	int err;

	rx4111_dev_info_check(spi);

	rx4111 = devm_kzalloc(&spi->dev, sizeof(struct rx4111_data), GFP_KERNEL);
	if (!rx4111)
		return -ENOMEM;

	rx4111->regmap = devm_regmap_init_spi(spi, &regmap_spi_config);
	if (IS_ERR(rx4111->regmap)) {
		dev_err(&spi->dev, "regmap init failed for rtc rx4111\n");
		return PTR_ERR(rx4111->regmap);
	}

	spi_set_drvdata(spi, rx4111);

	rx4111->rtc = devm_rtc_device_register(&spi->dev,RX4111_DRIVER_NAME, &rx4111_rtc_ops, THIS_MODULE);
	if (IS_ERR(rx4111->rtc))
		return PTR_ERR(rx4111->rtc);

	rx4111->irq = spi->irq;

	if (rx4111->irq > 0) {
		dev_info(&spi->dev, "IRQ %d supplied\n", spi->irq);
		err = devm_request_threaded_irq(&spi->dev, spi->irq, NULL,
				rx4111_irq_handler,IRQF_TRIGGER_FALLING | IRQF_ONESHOT,
				"rtc-rx4111", spi);
		if (err) {
			dev_err(&spi->dev, "unable to request IRQ\n");
			rx4111->irq = 0;
		} 
		else{
			rx4111_rtc_ops.read_alarm = rx4111_read_alarm;
			rx4111_rtc_ops.set_alarm = rx4111_set_alarm;
			rx4111_rtc_ops.alarm_irq_enable = rx4111_alarm_irq_enable;
		}

	}
	else
		dev_info(&spi->dev, "IRQ supplied fail\n");

	err = rx4111_init(rx4111);
	if (err)
		return err;

	rx4111->rtc->max_user_freq = 1;

	return 0;
}

static int rx4111_remove(struct spi_device *spi)
{
	return 0;
}

static const struct of_device_id rx4111_of_match[] = {
	{ .compatible = "epson,rtc-rx4111" },
	{ }
};
MODULE_DEVICE_TABLE(of, rx4111_of_match);

static struct spi_driver rx4111_driver = {
	.driver = {
		.name = RX4111_DRIVER_NAME,
		.of_match_table = of_match_ptr(rx4111_of_match),
	},
	.probe		= rx4111_probe,
	.remove		= rx4111_remove,
};

module_spi_driver(rx4111_driver);

MODULE_AUTHOR("Sam Kim <sam@varikorea.co.kr>");
MODULE_DESCRIPTION("RX-4111 CE RTC driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS("spi:rtc-rx4111");
